from typing import Set, Callable, List, Any, Dict

import networkx as nx
from evrouting.T import Node, SoC
from evrouting.graph_tools import (
    CONSUMPTION_KEY,
    DISTANCE_KEY,
    CHARGING_COEFFICIENT_KEY,
    consumption,
    distance,
    charging_cofficient
)

Path = List[Node]
DistFunction = Callable[[nx.Graph, Node, Node], Path]


def dijkstra(G: nx.Graph, u: Any, v: Any, weight: str = 'weight') -> Path:
    return nx.algorithms.shortest_path(G, u, v, weight=weight)


def fold_path(G: nx.Graph, path: Path, weight: str):
    return sum([G.edges[u, v][weight] for u, v in zip(path[:-1], path[1:])])


def insert_start_node(s: Node,
                      graph_core: nx.Graph,
                      gas_stations: Set[Node],
                      graph_extended: nx.DiGraph,
                      capacity: SoC,
                      initial_soc: SoC,
                      dist: DistFunction = dijkstra
                      ) -> nx.DiGraph:
    return graph_extended


def insert_final_node(t: Node,
                      graph_core: nx.Graph,
                      gas_stations: Set[Node],
                      graph_extended: nx.DiGraph,
                      capacity: SoC,
                      initial_soc: SoC,
                      dist: DistFunction = dijkstra
                      ) -> nx.DiGraph:
    return graph_extended


def contract_graph(G: nx.Graph, charging_stations: Set[Node], capacity: SoC,
                   dist: DistFunction = dijkstra) -> nx.Graph:
    """
    :param G: Original graph
    :param charging_stations: Charging stations
    :param capacity: Maximum battery capacity
    :param dist: Minimum distance function, necessary if G is not a fully
        connected Graph to calculate consumptions between charging stations.
    :returns: Graph only consisting of Charging Stations whose neighbours must
        be within the capacity U. If so, their edge has consumption and
        distance of the minimum path.
    """
    H: nx.Graph = nx.Graph()

    for cs in list(charging_stations):
        H.add_node(cs, **G.nodes[cs])
        # Iterate unvisited charging stations
        for n_cs in [n for n in charging_stations if (n, cs) not in H.edges]:
            min_path: Path = dist(G, cs, n_cs)
            consumption: SoC = fold_path(G, min_path, weight=CONSUMPTION_KEY)
            if consumption <= capacity:
                H.add_edge(
                    cs, n_cs,
                    **{
                        CONSUMPTION_KEY: consumption,
                        DISTANCE_KEY: fold_path(G, min_path, weight=DISTANCE_KEY)
                    }
                )
    return H


def get_possible_arriving_soc(G: nx.Graph, u: Node, capacity: SoC) -> List[SoC]:
    """
    :returns: All possible SoC  when arriving at node u, according to
        the optimal fuelling strategy.
    """
    possible_arriving_soc: Set[SoC] = {0}
    c_u = charging_cofficient(G, u)

    for n in G.neighbors(u):
        arriving_soc = capacity - consumption(G, u, n)
        if arriving_soc > 0 and charging_cofficient(G, n) < c_u and \
                arriving_soc not in possible_arriving_soc:
            possible_arriving_soc.add(arriving_soc)

    return list(possible_arriving_soc)


def state_graph(G: nx.Graph, capacity: SoC) -> nx.DiGraph:
    """Calculate Graph connecting (Node, Arrival SoC) states."""
    H: nx.DiGraph = nx.DiGraph()

    for u in G.nodes:
        c_u = charging_cofficient(G, u)
        for v in G.neighbors(u):
            w = consumption(G, u, v)
            if w <= capacity:
                for g in get_possible_arriving_soc(G, u, capacity):
                    c_v = charging_cofficient(G, v)
                    if c_v <= c_u and g < w:
                        weight = (w - g) * c_u + distance(G, u, v)
                        H.add_edge((u, g), (v, 0), weight=weight)
                    elif c_v > c_u:
                        weight = (capacity - g) * c_u + distance(G, u, v)
                        H.add_edge((u, g), (v, capacity - w), weight=weight)

    return H


def compose_result(G: nx.Graph, path: Path) -> dict:
    return {'trip_time': 0, 'path': path}


def shortest_path(G: nx.Graph, charging_stations: Set[Node], s: Node, t: Node,
                  initial_soc: SoC, final_soc: SoC, capacity: SoC) -> Dict:
    """
    Calculates shortest path using a generalized gas station algorithm.

    :param G: Input Graph
    :param s: Start Node identifier
    :param t: End Node identifier
    :param b_0: Start SoC
    :param b_t: End SoC
    :param U: Capacity
    :return:
    """
    S: nx.Graph = contract_graph(G, charging_stations, capacity)
    H = state_graph(S, capacity)

    insert_start_node(s, G, charging_stations, H, capacity, initial_soc)
    insert_final_node(t, G, charging_stations, H, capacity, final_soc)

    path: Path = dijkstra(H, (s, 0), (t, 0))

    return compose_result(H, path)
